# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Thermodynamic computations for metabolites


"""

from functools import reduce
from math import log, sqrt

from . import std
from ..utils.numerics import BIGM_THERMO

CPD_PROTON = 'cpd00067'

DEFAULT_VAL = BIGM_THERMO


class MetaboliteThermo:
    """
    A class representing the thermodynamic values of a enzyme

    :param dict metData: A dictionary containing the values for the
        enzyme, from the thermodynamic database
    :param float pH: The pH of the enzyme's compartment
    :param float ionicStr: The ionic strength of the enzyme's
        compartment
    :param temperature:
    :param min_ph:
    :param max_ph:
    :param debye_huckel_b:
    :param string thermo_unit: The unit used in `metData`'s values
    :param bool debug: *Optional* If set to ``True``, some debugging values
        will be printed. This is only useful for development or debugging
        purposes.

    .. note::

        The values are automatically computed on class creation. Usually you
        don't have to call any methods defined by this class, but only to access
        the attributes you need.

    The available attributes are :

    +-------------+-----------------------------------------------------------+
    | id          | The enzyme's seed_id                                   |
    +-------------+-----------------------------------------------------------+
    | pKa         | pKas of the enzyme                                    |
    +-------------+-----------------------------------------------------------+
    | error       | Error on the enzyme's thermodynamic data. This is the |
    |             | value from the thermodynamic database.                    |
    |             | Thermodynamic values will be computed only if this equals |
    |             | to 'Nil'.                                                 |
    +-------------+-----------------------------------------------------------+
    | deltaGf_std | Transformed Gibbs energy of formation of the enzyme,  |
    |             | in standard conditions (value from the the thermodynamic  |
    |             | database).                                                |
    +-------------+-----------------------------------------------------------+
    | deltaGf_err | Error on the transformed Gibbs energy of formation of the |
    |             | enzyme,in standard conditions (value from the the     |
    |             | thermodynamic database).                                  |
    +-------------+-----------------------------------------------------------+
    | mass        | Mass of the enzyme (value from the thermodynamic      |
    |             | database).                                                |
    +-------------+-----------------------------------------------------------+
    | nH_std      | Number of protons of the enzyme, in standard          |
    |             | conditions (extracted from the thermodynamic database).   |
    +-------------+-----------------------------------------------------------+
    | struct_cues | The cues that make the structure of the enzyme.       |
    +-------------+-----------------------------------------------------------+
    | charge_std  | The charge (mV) of the enzyme, in standard conditions.|
    |             | This value is extracted from the thermodynamic database.  |
    +-------------+-----------------------------------------------------------+
    | pH          | The pH of the enzyme's compartment                    |
    +-------------+-----------------------------------------------------------+
    | ionicStr    | The ionic strength of the enzyme's compartment        |
    +-------------+-----------------------------------------------------------+
    | deltaGf_tr  |  Transformed Gibbs energy of formation of specie with     |
    |             |  given pH and ionic strength using formula given by       |
    |             |  Goldberg and Tewari, 1991                                |
    +-------------+-----------------------------------------------------------+

    Since the reactions expose similar values through a dictionnary, it is
    better to access the attributes aforementionned of this class as if it was
    a dictionnary : ``enzyme.thermo['pH']``.

    """

    def __init__(self, metData, pH, ionicStr, temperature=std.TEMPERATURE_0,
                 min_ph=std.MIN_PH, max_ph=std.MAX_PH,
                 debye_huckel_b=std.DEBYE_HUCKEL_B_0, thermo_unit='kJ/mol',
                 debug=False):
        # Store all the data we got
        """

        :param metData:
        :param pH:
        :param ionicStr:
        :param temperature:
        :param min_ph:
        :param max_ph:
        :param debye_huckel_b:
        :param thermo_unit:
        :param debug:
        """
        self.debug = debug
        if self.debug and metData is not None:
            print("Debugging thermo computations for enzyme : {} ({})"
                  .format(metData['name'], metData['id'])
                  )

        self.pH = pH
        self.ionicStr = ionicStr
        self.Debye_Huckel_B = debye_huckel_b

        # Compute internal values to adapt the the thermo_unit provided
        if thermo_unit == "kJ/mol":
            GAS_CONSTANT = 8.314472 / 1000  # kJ/(K mol)
            self.Adjustment = 1
        else:
            GAS_CONSTANT = 1.9858775 / 1000  # Kcal/(K mol)
            self.Adjustment = 4.184

        TEMPERATURE = temperature

        self.RT = GAS_CONSTANT * TEMPERATURE

        # CONSTANTS
        self.MAX_pH = max_ph
        self.MIN_pH = min_ph

        # Store metData values
        self.id = None if metData == None else metData['id']
        self.pKa = [] if metData == None else metData['pKa']
        self.error = None if metData == None else metData['error']
        self.deltaGf_std = DEFAULT_VAL if metData == None else metData[
            'deltaGf_std']
        self.deltaGf_err = DEFAULT_VAL if metData == None else metData[
            'deltaGf_err']
        self.mass = DEFAULT_VAL if metData == None else metData['mass_std']
        self.nH_std = None if metData == None else metData['nH_std']
        self.struct_cues = [] if metData == None else metData['struct_cues']
        self.charge_std = DEFAULT_VAL if metData == None else metData['charge_std']
        self.struct_cues = None if metData == None else metData['struct_cues']

        # Compute deltaGf_tr if possible
        self.deltaGf_tr = DEFAULT_VAL if metData == None else self.calcDGis()

        self.__dict__ = {
            'id': self.id,
            'pKa': self.pKa,
            'error': self.error,
            'deltaGf_std': self.deltaGf_std,
            'deltaGf_err': self.deltaGf_err,
            'mass': self.mass,
            'nH_std': self.nH_std,
            'charge_std': self.charge_std,
            'struct_cues': self.struct_cues,
            'deltaGf_tr': self.deltaGf_tr,
            'pH': self.pH,
            'ionicStr': self.ionicStr,
        }

    # Various methods to have a dictionnary-like behavior, for consistency with
    # the reactions' thermo attribute
    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def __cmp__(self, dict_):
        return cmp(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __unicode__(self):
        return unicode(repr(self.__dict__))

    def calcDGis(self):
        """ Calculate the transformed Gibbs energy of formation of specie with
        given pH and ionic strength using formula given by Goldberg and Tewari,
        1991

        Equation 4.5-6 in Alberty's book

        :returns: DG_is for the enzyme
        :rtype: float

        """

        if self.debug:
            print("Computing DGis...")

        # Special case for protons...
        if self.id == CPD_PROTON:
            if self.debug:
                print("Found proton")
            return -self.RT * log(10 ** -self.pH)

        # Error too big
        if self.error != 'Nil' or self.deltaGf_std > 9 * 10 ** 6:
            if self.debug:
                print("Error or deltaGf_std too big")
            return 10 ** 7

        # Compute the value
        P = self.calc_potential()
        DGis = self.calcDGsp() - self.RT * log(P)

        if self.debug:
            print("Found DGis = " + str(DGis))

        return DGis

    def calcDGsp(self):
        """ Calculate the transformed Gibbs energy of formation of specie with
        given pH and ionic strength using formula given by Goldberg and Tewari,
        1991

        Equation 4.4-10 in Alberty's book

        :returns: DG_sp for the enzyme
        :rtype: float


        """
        (deltaGo, charge, nH) = self.calcDGspA()
        zsq = charge ** 2
        I = self.ionicStr
        term1 = (nH * self.RT * log(10 ** -self.pH))
        term2 = (2.91482
                 * (zsq - nH)
                 * sqrt(I)
                 / (1 + self.Debye_Huckel_B * sqrt(I))
                 ) / self.Adjustment

        return deltaGo - (term1 + term2)

    def calc_potential(self):
        """ Calculate the binding polynomial of a specie, with pK values

        :returns: The potential of the enzyme
        :rtype: float

        """
        if self.debug:
            print("Computing P...")

        # Init some values used here...
        prod_denom = 1
        p = 1
        pka_values = self.get_pka()

        if self.debug:
            print("pKas used : " + str(pka_values))

        # Make the computation...
        if len(pka_values) > 0:
            if min(pka_values) <= self.MAX_pH:
                for i, this_pka in enumerate(pka_values):
                    numerator = 10 ** (-(i + 1) * self.pH)
                    K = 10 ** (-this_pka)
                    denominator = prod_denom * K
                    prod_denom = denominator
                    if self.debug:
                        print("Adding "
                              + str(numerator / denominator)
                              + " to P")

                    p += numerator / denominator

        return p

    def get_pka(self):
        """ Get the pKas of the enzyme

        :returns: The pKas of the enzyme
        :rtype: list(float)

        """
        if self.debug:
            print("Getting the list of pKas...")
        (deltaGspA, charge, sp_nH) = self.calcDGspA()

        pka_list = self.pKa

        pka_values = [None] * len(pka_list)

        # Get only useful pKas
        pka_list = [x for x in pka_list if self.MIN_pH < x < self.MAX_pH]
        # Sort the list
        pka_list.sort(reverse=True)

        j = 0

        if len(pka_list) > 1:
            for i,this_pka in enumerate(pka_list):
                sigmanusq = 1 + (charge + i) ** 2 - (charge + i - 1) ** 2
                if self.MAX_pH > pka_list[i] > self.MIN_pH:
                    pka_values[j] = self._calc_pka(this_pka,sigmanusq)

                    if self.debug:
                        print("Added to pKas : " + str(pka_values[j]))

                    j += 1

        elif len(pka_list) == 1:
            if self.debug:
                print("Only one pKa to add")

            sigmanusq = 2 * charge
            pka_values[j] = self._calc_pka(pka_list[j],sigmanusq)

        # Only return useful values
        pka_values = [x for x in pka_values if x is not None]

        if self.debug:
            print("Filtered pKa values : " + str(pka_values))

        pka_values.sort(reverse=True)

        return pka_values

    def _calc_pka(self, pka,sigmanusq):
        lnkzero = log(10 ** -pka)
        pka_value = -(
            lnkzero - sigmanusq * (std.DEBYE_HUCKEL_A * log(10) * sqrt(self.ionicStr)) / (
            1 + self.Debye_Huckel_B * sqrt(self.ionicStr))) / log(10)
        return pka_value

    def calcDGspA(self):
        """ Calculates deltaGf, charge and nH of the specie when it is at least
        protonated state based on MFAToolkit compound data for the pKa values
        within the range considered (MIN_pH to MAX_pH).

        These values are used as the starting point for Alberty's calculations.

        :returns: deltaGspA, sp_charge and sp_nH
        :rtype: tuple(float, float, int)

        """
        if self.debug:
            print('Computing DGspA()...')

        # Case of the proton
        if self.id == CPD_PROTON:
            if self.debug:
                print('Proton found, returning standard values')
            # we do not adjust for proton so just return the values
            deltaGspA = self.deltaGf_std
            sp_charge = self.charge_std
            sp_nH = self.nH_std
            return (deltaGspA, sp_charge, sp_nH)

        pka_list = self.pKa

        acceptedpKas = [x for x in pka_list if (
            self.MIN_pH < x < self.MAX_pH
        )]

        # No pKas found
        if len(acceptedpKas) == 0:
            if self.debug:
                print('No pKas found, returning standard values')
            deltaGspA = self.deltaGf_std
            sp_charge = self.charge_std
            sp_nH = self.nH_std
            return (deltaGspA, sp_charge, sp_nH)

        # if compound is multiprotic
        # discard pKa values above MAX_pH

        pka_list = [x for x in pka_list if x < self.MAX_pH]
        charge_adj = reduce(lambda count, x: count + 1 if x < 7 else count,
                            acceptedpKas,
                            0)
        sp_charge = -len(pka_list)

        pKs = []
        sp_deltaGf = 0
        sp_nH = 0

        # if the ReactionDB compound is already at the right state, i.e. most
        # negative state we can just return their values too after adjusting
        # for the ionic strength

        if self.charge_std == sp_charge:
            if self.debug:
                print('charge found is equal to standard charge, '
                      + 'returning default values')
            deltaGspA = self.deltaGf_std
            sp_charge = self.charge_std
            sp_nH = self.nH_std
            return (deltaGspA, sp_charge, sp_nH)

        charge = self.charge_std

        num_iter = charge - sp_charge

        sp_nH = self.nH_std - num_iter
        deltaGspA = self.deltaGf_std

        npKa = len(pka_list)

        if npKa - num_iter == -1:
            start = 0  # NOTE : In Python, lists begin at 0
        else:
            start = npKa - num_iter

        pKs = pka_list[start:]

        if self.debug:
            print("Accepted pKs : ", pKs)

        if self.debug:
            print(pka_list, start, pKs)

        for j,_ in enumerate(pKs):
            deltaGspA -= self.RT * log(10 ** -pKs[j])

        if self.debug:
            print("Found deltaGspA : " + str(deltaGspA))

        return (deltaGspA, sp_charge, sp_nH)


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Thermodynamic computations for reactions


"""
from functools import reduce
from math import log, sqrt

from . import std
from .utils import find_transported_mets
from .metabolite import CPD_PROTON

###################
# REACTIONS TOOLS #
###################



def calcDGtpt_rhs(reaction, compartmentsData, thermo_units):
    """ Calculates the RHS of the deltaG constraint, i.e. the sum of the
    non-concentration terms

    :param cobra.thermo.reaction.Reaction reaction: The reaction to compute the
        data for
    :param dict(float) compartmentsData: Data of the compartments of the cobra_model
    :param str thermo_units: The thermodynamic database of the cobra_model

    :returns: deltaG_tpt and the breakdown of deltaG_tpt
    :rtype: tuple(float, dict(float))

    Example:
        ATP Synthase reaction::

            reaction = cpd00008 + 4 cpd00067 + cpd00009 <=> cpd00002 + 3 cpd00067 + cpd00001
            compartments =  'c'       'e'        'c'           'c'         'c'         'c'

    If there are any metabolites with unknown energies then returns
        ``(0, None)``.

    """

    # Compute our constants in accordance with the thermoDB
    if thermo_units == "kJ/mol":
        GAS_CONSTANT = 8.314472 / 1000  # kJ/(K mol)
        faraday_const = 96.485  # kJ/eV
    else:
        GAS_CONSTANT = 1.9858775 / 1000  # Kcal/(K mol)
        faraday_const = 23.061  # kcal/eV

    TEMPERATURE = 298.15  # K

    RT = GAS_CONSTANT * TEMPERATURE

    if reduce(lambda count, met: (
                count + (1 if met.thermo.deltaGf_tr > 10 ** 6 else count)),
              reaction.metabolites,
              0) > 1:
        return (0, None)

    sum_deltaGFis_trans = 0
    sum_stoich_NH = 0
    RT_sum_H_LC_tpt = 0  # to include the differential proton concentration
    # effects if protons are transported

    transportedMets = find_transported_mets(reaction)
    compartments = {'reactant': [], 'product': []}

    for seed_id in transportedMets:
        for metType in ['reactant', 'product']:
            if seed_id != 'cpd00001':
                met = transportedMets[seed_id][metType]
                pH_comp = met.thermo.pH
                ionicStr_comp = met.thermo.ionicStr

                deltaGfsp = met.thermo.deltaGf_tr

                compartments[metType].append(met.compartment)
                sum_stoich_NH += ((1 if metType == 'product' else -1)
                                  * transportedMets[seed_id]['coeff']
                                  * met.thermo.nH_std
                                  * RT
                                  * log(10 ** -pH_comp))
                sum_deltaGFis_trans += ((1 if metType == 'product' else -1)
                                        * transportedMets[seed_id]['coeff']
                                        * deltaGfsp)
            else:
                compartments[metType].append('')

            if seed_id == CPD_PROTON:
                met = transportedMets[seed_id][metType]
                pH_comp = met.thermo.pH
                RT_sum_H_LC_tpt += ((1 if metType == 'product' else -1)
                                    * RT
                                    * transportedMets[seed_id]['coeff']
                                    * log(10 ** -pH_comp))

    # calculate the transport of any ions
    # membrane potential is always defined as inside - outside
    # we should take the larger stoich of the transported compound
    sum_F_memP_charge = 0

    for seed_id in transportedMets:
        if seed_id != 'cpd00001':
            out_comp = transportedMets[seed_id]['reactant'].compartment
            in_comp = transportedMets[seed_id]['product'].compartment
            mem_pot = compartmentsData[out_comp]['membranePot'][in_comp]
            charge = transportedMets[seed_id]['reactant'].thermo.charge_std
            # Equal to the product's one
            sum_F_memP_charge += (faraday_const
                                  * (mem_pot / 1000.)
                                  * transportedMets[seed_id]['coeff']
                                  * charge)

    deltaG = 0

    for met in reaction.metabolites:
        if CPD_PROTON != met.annotation['seed_id']:
            deltaG += reaction.metabolites[met] * met.thermo.deltaGf_tr

    sum_deltaGFis = 0

    # lastly we calculate the deltaG of the chemical reaction if any
    # but we do not add this part to the rhs as it would be included in the
    # potential energy of the enzyme


    final_coeffs = reaction.metabolites.copy()

    for seed_id in transportedMets:
        for metType in ['reactant', 'product']:
            final_coeffs[transportedMets[seed_id][metType]] -= (
                (1 if metType == 'product' else -1)
                * transportedMets[seed_id]['coeff'])

    for met in final_coeffs:
        if final_coeffs[met] != 0 and met.annotation['seed_id'] != CPD_PROTON:

            met_deltaGis = met.thermo.deltaGf_tr
            sum_deltaGFis += final_coeffs[met] * met_deltaGis

    # Sum all the parts
    DG_trans_RHS = (sum_stoich_NH
                    + sum_F_memP_charge
                    + sum_deltaGFis_trans
                    + RT_sum_H_LC_tpt
                    + sum_deltaGFis)

    breakdown = {
        'sum_deltaGFis': sum_deltaGFis,
        'sum_stoich_NH': sum_stoich_NH,
        'sum_F_memP_charge': sum_F_memP_charge,
        'sum_deltaGFis_trans': sum_deltaGFis_trans,
        'RT_sum_H_LC_tpt': RT_sum_H_LC_tpt
    }

    return (DG_trans_RHS, breakdown)


def calcDGR_cues(reaction, reaction_cues_data):
    """ Calculates the deltaG reaction and error of the reaction using the
    constituent structural cues changes and returns also the error if any.

    :param cobra.thermo.reaction.Reaction reaction: The reaction to compute
        deltaG for
    :param dict reaction_cues_data:

    :returns: deltaGR, error on deltaGR, the cues in the reaction (keys of the
        dictionnary) and their indices (values of the dictionnary),
        and the error code if any.

        If everything went right, the error code is an empty string

    :rtype: tuple(float, float, dict(float), str)

    """

    deltaGR = 0
    deltaGR_err = 0
    cues = {}
    error = ''

    # First we should check if all the reactants are in terms of compound IDs
    for reactant in reaction.metabolites:
        if len(reactant.thermo.struct_cues) == 0:
            return (10 ** 7, 10 ** 7, '', 'UNKNOWN_GROUPS')
        (deltaGF, deltaGFerr, cpd_cues) = calcDGF_cues(
            reactant.thermo.struct_cues,
            reaction_cues_data)
        for cue in cpd_cues:
            if cue in cues:
                cues[cue] += reaction.metabolites[reactant] * cpd_cues[cue]
            else:
                cues[cue] = reaction.metabolites[reactant] * cpd_cues[cue]

    for cue in cues:
        deltaGR += cues[cue] * reaction_cues_data[cue]['energy']
        deltaGR_err += (cues[cue] * reaction_cues_data[cue]['error']) ** 2

    deltaGR_err = sqrt(deltaGR_err)

    return (deltaGR, deltaGR_err, cues, error)


def calcDGF_cues(cues, reaction_cues_data):
    """ Calculates the deltaG formation and error of the compound using its
    constituent structural cues.

    :param list(str) cues: A list of cues' names
    :param dict reaction_cues_data:

    :returns: deltaG formation, the error on deltaG formation, and a dictionnary
        with the cues' names as key and their coefficient as value
    :rtype: tuple(float, float, dict(float)).

    """
    deltaGF = 0
    deltaGF_err = 0
    finalcues = {}

    for cue in cues:
        if cue in finalcues:
            finalcues[cue] += cues[cue]
        else:
            finalcues[cue] = cues[cue]

        deltaGF += reaction_cues_data[cue]['energy'] * cues[cue]
        deltaGF_err += (reaction_cues_data[cue]['error'] * cues[cue]) ** 2

    deltaGF_err = sqrt(deltaGF_err)

    return (deltaGF, deltaGF_err, finalcues)


def get_debye_huckel_b(T):
    """
    The Debye-Huckel A and B do depend on the temperature
    As for now though they are returned as a constant (value at 298.15K)

    :param T: Temperature in Kelvin
    :return: Debye_Huckel_B
    """
    return std.DEBYE_HUCKEL_B_0


"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Input/Output tools to import or export pytfa models


"""

import pickle
import zlib
import numpy as np
import re

from cobra import Model, Reaction, Metabolite
from cobra.io import load_matlab_model
from scipy.io import loadmat, savemat

from ..utils.numerics import BIGM_DG

from warnings import warn

try:
    from scipy.sparse import dok_matrix, lil_matrix
except ImportError:
    dok_matrix, lil_matrix = None, None


def import_matlab_model(path, variable_name=None):
    """ Convert at matlab cobra_model to a pyTFA cobra_model, with Thermodynamic values

    :param variable_name:
    :param string path: The path of the file to import

    :returns: The converted cobra_model
    :rtype: cobra.thermo.model.Model

    """
    # We're going to import the Matlab cobra_model and convert it to COBApy
    mat_data = loadmat(path)

    if variable_name:
        mat_model = mat_data[variable_name][0, 0]
    else:
        meta_vars = {"__globals__", "__header__", "__version__"}
        possible_keys = sorted(i for i in mat_data if i not in meta_vars)

        if len(possible_keys) == 1:
            variable_name = possible_keys[0]
        else:
            raise Exception(
                'Need to specify a variable name if several ' + 'variables are contained int the mat file')

        mat_model = mat_data[variable_name][0, 0]

    # cobra_model = Model(mat_model['description'][0])
    cobra_model = load_matlab_model(path)

    cobra_model.description = mat_model['description'][0]
    cobra_model.name = mat_model['description'][0]

    metabolites = cobra_model.metabolites

    ## METABOLITES
    # In the Matlab cobra_model, the corresponding components are :
    # * mets = Identifiers
    # * metNames = Names
    # * metFormulas = formulas
    # * metCompSymbol = compartments

    # def read_mat_model(mat_struct, field_name, index):
    #     try:
    #         return mat_struct[field_name][index,0][0]
    #     except IndexError:
    #         return None

    # metabolites = [Metabolite( read_mat_model(mat_model,'mets',i),
    #     formula=read_mat_model(mat_model,'metFormulas',i),
    #     name=read_mat_model(mat_model,'metNames',i),
    #     compartment=read_mat_model(mat_model,'metCompSymbol',i))
    #                for i in range(len(mat_model['metNames']))]

    # Get the metSEEDID
    seed_id = mat_model['metSEEDID']
    for i, _ in enumerate(metabolites):
        metabolites[i].annotation = {"seed_id": seed_id[i, 0][0]}

    # ## REACTIONS
    # # In the Matlab cobra_model, the corresponding components are :
    # # * rxns = Names
    # # * rev = Reversibility (not used, see https://cobrapy.readthedocs.io/en/0.5.11/faq.html#How-do-I-change-the-reversibility-of-a-Reaction?)
    # # * lb = Lower bounds
    # # * ub = Upper bounds
    # # * subSystems = subsystem names
    # # * S : Reactions matrix
    # #       1 line = 1 metabolite
    # #       1 column = 1 reaction
    # # * c : Objective coefficient
    # # * genes : Genes names
    # # * rules : Genes rules

    # # Some utilities for gene generation rules (convert the IDs to the names of the genes)
    # gene_pattern = re.compile(r'x\([0-9]+\)')

    # def gene_id_to_name(match):
    #     id_ = int(match.group()[2:-1])
    #     # /!\ These are indexed from 1, while python indexes from 0
    #     return mat_model['genes'][id_ - 1, 0][0]

    # # Add each reaction
    # for i in range(mat_model['S'].shape[1]):
    #     # Name of the reaction
    #     reaction = Reaction(str(mat_model['rxns'][i, 0][0]))

    #     # Add the reaction to the cobra_model
    #     cobra_model.add_reactions([reaction])

    #     # NOTE : The str() conversion above is needed, otherwise the CPLEX solver
    #     # does not work : "Invalid matrix input type --"

    #     # Reaction description
    #     # reaction.name not set

    #     # Subsystem (only if set in the Matlab cobra_model)
    #     if (len(mat_model['subSystems'][i, 0])):
    #         reaction.subsystem = mat_model['subSystems'][i, 0][0]

    #     # Lower bound
    #     reaction.lower_bound = float(mat_model['lb'][i, 0])
    #     # Upper bound
    #     reaction.upper_bound = float(mat_model['ub'][i, 0])

    #     # Objective coefficient
    #     reaction.objective_coefficient = float(mat_model['c'][i, 0])

    #     # Metabolites
    #     react_mets = {}
    #     # Iterate over each metabolite and see if it is part of the reaction
    #     # (stoechiomectric coefficient not equal to 0)
    #     for j, _ in enumerate(metabolites):
    #         if (mat_model['S'][j, i] != 0):
    #             react_mets[metabolites[j]] = mat_model['S'][j, i]

    #     reaction.add_metabolites(react_mets)

    #     # Genes
    #     try:
    #         if len(mat_model['rules'][i, 0]):
    #             rule = mat_model['rules'][i, 0][0]
    #             # Call the regex magic to convert IDs to gene names
    #             rule = gene_pattern.sub(gene_id_to_name, rule)

    #             # Add the data to the reaction
    #             reaction.gene_reaction_rule = rule
    #     except ValueError:
    #         pass
    #     except IndexError:
    #         # The gene number is higher than the length of the gene list
    #         warn('The gene reaction rule {} appears to be misaligned with '
    #              'the gene list'.format(rule))


    Compartments = dict()
    CompartmentDB = mat_model['CompartmentData'][0]

    num_comp = len(CompartmentDB['pH'][0][0])
    comps = [{} for i in range(num_comp)]

    for i in range(num_comp):
        comp = comps[i]
        comp['pH'] = CompartmentDB['pH'][0][0][i]
        comp['ionicStr'] = CompartmentDB['ionicStr'][0][0][i]
        comp['symbol'] = CompartmentDB['compSymbolList'][0][0, i][0]
        comp['name'] = CompartmentDB['compNameList'][0][0, i][0]
        comp['c_max'] = CompartmentDB['compMaxConc'][0][0][i]
        comp['c_min'] = CompartmentDB['compMinConc'][0][0][i]

        # Register our compartment by its name
        if comp['symbol'] in Compartments:
            raise Exception("Duplicate compartment ID : " + comp['symbol'])

        Compartments[comp['symbol']] = comp

    # We need to iterate first once to set the names of the compartments, so that we have the keys for our dictionnary...
    for i in range(num_comp):
        comp = comps[i]
        comp['membranePot'] = {}
        for j in range(num_comp):
            comp['membranePot'][comps[j]['symbol']] = float(CompartmentDB['membranePot'][0][i, j])

    cobra_model.compartments = Compartments

    recover_compartments(cobra_model, list(Compartments.keys()))

    return cobra_model

def recover_compartments(model, compartments_list):
    for met in model.metabolites:
        if met.compartment is None:
            comp_candidate = met.id.split('_')[-1]
            if comp_candidate in compartments_list:
                met.compartment = comp_candidate
            else:
                model.logger.info('No compartment data found for metabolite {}'.format(met.id))
        else:
            continue


def write_matlab_model(tmodel, path, varname='tmodel'):
    """
    Writes the Thermo Model to a Matlab-compatible structure

    :param varname:
    :param tmodel:
    :param path:
    :return: None
    """

    from cobra.io.mat import create_mat_dict

    mat = create_mat_dict(tmodel)
    mat.update(create_thermo_dict(tmodel))
    mat.update(create_problem_dict(tmodel))
    savemat(path, {varname: mat}, appendmat=True,
                     oned_as="column")


def create_thermo_dict(tmodel):
    """
    Dumps the thermodynamic information in a mat-compatible dictionary
    (similar to the output of cobra.io.mat.create_mat_dict)

    :param tmodel: pytfa.thermo.tmodel.ThermoModel
    :return: dict object
    """

    mat = {}

    met_map = {
        'metSEEDID':('id',''),
        'metDeltaGFstd':('deltaGf_std',BIGM_DG),
        'metDeltaGFerr':('deltaGf_err',BIGM_DG),
        'metMass':('mass',BIGM_DG),
        'metCharge':('charge_std',BIGM_DG),
        'metDelGFtr':('deltaGf_tr',BIGM_DG),
        'metCompSymbol':('compartment','')
    }

    for column,(key,default_value) in met_map.items():
        the_data = np.full(len(tmodel.metabolites),default_value)

        for e,the_met in enumerate(tmodel.metabolites):
            try:
                the_value = the_met.thermo[key]
                the_data[e] = the_value
            except KeyError:
                continue

        if column == 'metSEEDID':
            mat[column] = np.array(the_data, dtype=np.object)
        else:
            mat[column] = np.array(the_data)


    rxn_map = {
        'rxnDeltaGR':('DeltaGR',BIGM_DG),
        'rxnDeltaGRerr':('DeltaGRerr',BIGM_DG),
        'rxnThermo':('computed',None),
        'isTrans':('isTrans',None),
    }

    for column,(key,default_value) in rxn_map.items():
        the_data = np.full(len(tmodel.reactions),default_value)

        for e,the_rxn in enumerate(tmodel.reactions):
            try:
                the_value = the_rxn.thermo[key]
                the_data[e] = the_value
            except KeyError:
                continue

        mat[column] = np.array(the_data)

    # Adding compartment data
    CompartmentDB = {}

    compartments = [v for v in tmodel.compartments.values()]
    CompartmentDB['pH'] = np.array([x['pH'] for x in compartments])
    CompartmentDB['ionicStr'] = np.array([x['ionicStr'] for x in compartments])
    CompartmentDB['compMaxConc'] = np.array([x['c_max'] for x in compartments])
    CompartmentDB['compMinConc'] = np.array([x['c_min'] for x in compartments])

    #Write symbols and names in collumn cell arrays
    CompartmentDB['compSymbolList'] = np.zeros((1, len(compartments)), dtype=np.object)
    CompartmentDB['compNameList'] =  np.zeros((1, len(compartments)), dtype=np.object)

    mat_to_python_string = [('compSymbolList', 'symbol'),
                            ('compNameList', 'name')]

    for i, _ in enumerate(compartments):
        for mat_name, py_name in mat_to_python_string:
            CompartmentDB[mat_name][0, i] = compartments[i][py_name]


    # The membrane potential is an NxN matrix in the matlab format
    membrane_pot = np.zeros((len(compartments),len(compartments)))
    for i, _ in enumerate(compartments):
        for j, _ in enumerate(compartments):
            symbol_j = compartments[j]['symbol']
            membrane_pot[i,j] = compartments[i]['membranePot'][symbol_j]

    CompartmentDB['membranePot'] = membrane_pot

    mat['CompartmentData'] = CompartmentDB

    return mat


def varnames2matlab(name, tmodel):
    """
    Transforms reaction variable pairs from `('ACALD','ACALD_reverse_xxxxx')` to
    `('F_ACALD','B_ACALD')`   if it is a reaction, else leaves is as is

    :return:
    """

    reverse_regex = re.compile(r'(.+_reverse)_[a-f0-9]{5}')

    new_name = name

    if new_name in tmodel.reactions:
        new_name = 'F_' + new_name
    else:
        test = reverse_regex.match(new_name)
        if test:
            new_name = 'R_' + test.groups()[0]

    return new_name



def create_problem_dict(tmodel):
    """
    Dumps the the MILP formulation for TFA in a mat-compatible dictionary
    (similar to the output of cobra.io.mat.create_mat_dict)

    :param tmodel: pytfa.thermo.tmodel.ThermoModel
    :ret
    """

    vartype_map = {
        'continuous':'C',
        'binary':'B',
        'integer':'B' # FIXME Waiting for optlang fix on binary vars
    }

    mat= {}

    mat['A'] = create_generalized_matrix(tmodel)

    # Variables

    mat['var_lb'] = np.array([x.lb for x in tmodel.variables]) * 1.
    mat['var_ub'] = np.array([x.ub for x in tmodel.variables]) * 1.
    vname = np.full(len(tmodel.variables),'', dtype=np.object)
    vtype = np.full(len(tmodel.variables),'', dtype=np.object)

    for e, this_var in enumerate(tmodel.variables):
        this_name = this_var.name
        new_name = varnames2matlab(this_name, tmodel)

        vname[e] = new_name
        vtype[e] = vartype_map[this_var.type]

    mat['varNames'] = vname
    mat['vartypes'] = vtype

    obj = tmodel.objective
    mat['objtype'] = -1 if tmodel.objective.direction.startswith('max') else 1
    mat['f'] = np.full(len(tmodel.variables), 0, dtype = np.double) * 1.0

    for this_var,this_coeff in obj.get_linear_coefficients(obj.variables).items():
        matlab_name = varnames2matlab(this_var.name, tmodel)
        mat['f'][np.where(mat['varNames']==matlab_name)[0][0]] = this_coeff

    # Constraints

    rhs = np.empty(len(tmodel.constraints))
    ctype = np.full(len(tmodel.constraints),'', dtype=np.object)
    cname = np.full(len(tmodel.constraints),'', dtype=np.object)

    for e,this_cons in enumerate(tmodel.constraints):
        if   this_cons.lb is None and this_cons.ub is not None:
            this_type = '<'
            this_rhs = this_cons.ub
        elif this_cons.ub is None and this_cons.lb is not None:
            this_type = '>'
            this_rhs = this_cons.lb
        elif this_cons.ub is not None and this_cons.lb is not None:
            if this_cons.lb == this_cons.ub:
                this_type = '='
                this_rhs = this_cons.ub
            else:
                raise(Exception('Matlab formulation does not support '
                                'constraints under the form: lb<x<ub'))
        else:
            raise(ValueError('Constraint type not recognized'))

        rhs[e] = this_rhs
        ctype[e] = this_type
        cname[e] = this_var.name

    mat['rhs'] = rhs *1
    mat['constraintType'] = ctype
    mat['constraintNames'] = cname

    return mat


def create_generalized_matrix(tmodel, array_type = 'dense'):
    """
    Returns the generalized stoichiomatric matrix used for TFA

    :param array_type:
    :param tmodel: pytfa.ThermoModel

    :returns: matrix.
    """

    if array_type not in ('DataFrame', 'dense') and not dok_matrix:
        raise ValueError('Sparse matrices require scipy')

    dtype = np.float64

    array_constructor = {'dense': np.zeros, 'dok': dok_matrix,
        'lil': lil_matrix, 'DataFrame': np.zeros, }

    n_constraints = len(tmodel.constraints)
    n_variables = len(tmodel.variables)
    array = array_constructor[array_type]((n_constraints, n_variables),
                                          dtype=dtype)

    c_ind = {x:e for e,x in enumerate(tmodel.constraints)}
    v_ind = {x:e for e,x in enumerate(tmodel.variables)}

    for this_cons in tmodel.constraints:
        var_coeff_dict = this_cons.get_linear_coefficients(this_cons.variables)

        for this_var,coeff in var_coeff_dict.items():
            array[c_ind[this_cons], v_ind[this_var]] = coeff

    if array_type == 'DataFrame':
        metabolite_ids = [met.id for met in tmodel.constraints]
        reaction_ids = [rxn.id for rxn in tmodel.variables]
        return pd.DataFrame(array, index=metabolite_ids, columns=reaction_ids)

    else:
        return array


def load_thermoDB(path):
    """ Load a thermodynamic database

    :param string path: The path of the file to load
    :returns: The thermodynamic database
    :rtype: dict

    """
    with open(path, 'rb') as file:
        ReactionDB = pickle.loads(zlib.decompress(file.read()))

    return ReactionDB


def printLP(model):
    """ Print the LP file corresponding to the cobra_model

    :param cobra.thermo.model.Model model: The cobra_model to output the LP file for

    :returns: The content of the LP file
    :rtype: str

    Usually, you pass the result of this function to :py:func:`file.write` to
    write it on disk. If you prefer, you can use :func:`pytfa.io.writeLP` to
    write the result directly to a file.

    """
    # It's much more efficient to allocate the file in memory then write it as
    # a whole to the disk, rather than making multiple disk calls
    res = ''

    # Write the problem name
    res += '\\Problem name: {}_LP\n\nMaximize\n obj: '.format(model.description)

    # Write the objective
    for rxn in model.reactions:
        if rxn.objective_coefficient != 0:
            if rxn.objective_coefficient != 1:
                res += rxn.objective_coefficient + ' '
            res += rxn.id

    # Constraints
    res += '\nSubject To\n'

    for cons in model.constraints:
        # Name of the constraint :
        res += cons.name + ': '

        # Write the lower bound if applicable...
        if cons.lb != cons.ub:
            if cons.lb is not None:
                res += str(cons.lb) + ' < '

        # Write the bound
        res += str(cons.expression)

        # Write the upper bound
        if cons.lb == cons.ub:
            res += ' = ' + str(cons.ub)
        elif cons.ub is not None:
            res += ' < ' + str(cons.ub)

        # Next line
        res += '\n'

    # Variables
    res += 'Bounds\n'

    for var in model.variables:
        # optlang already does the hard job for us, yay !
        res += str(var) + '\n'

    # Binrary constraints
    res += 'Binaries\n'

    # Number of variables on the current line we're writing
    count = 1

    for var in model.variables:
        # FIXME using the integer trick to be able to constraint binary variables
        # if var.type=='binary':
        if var.type in ['binary', 'integer']:
            res += var.name + '\t'
            count += 1
            # Print at most 7 variables per line
            if count == 7:
                res += '\n'
                count = 1

    # Done !
    res += '\nEnd'

    return res


def writeLP(model, path=None):
    """ Write the LP file of the specified cobra_model to the file indicated by path.

    :param cobra.thermo.model.Model model: The COBRApy cobra_model to write the LP file
        for
    :param string path: `Optional` The path of the file to be written. If not
        specified, the name of the COBRApy cobra_model will be used.


    """
    if not path:
        path = model.description + '.lp'

    with open(path, 'w') as file:
        file.write(printLP(model))



"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Sampling wrappers for pytfa models


"""

import numpy as np
from sympy.core.singleton import S
from time import  time
from  cobra.sampling import OptGPSampler, ACHRSampler, HRSampler,\
                                            shared_np_array
from optlang.interface import OPTIMAL

class GeneralizedHRSampler(HRSampler):

    def __init__(self, model, thinning,  nproj=None, seed=None):
        """
        Adapted from cobra.flux_analysis.sampling.py
        _________________________________________

        Initialize a new sampler object.
        """
        # This currently has to be done to reset the solver basis which is
        # required to get deterministic warmup point generation
        # (in turn required for a working `seed` arg)
        HRSampler.__init__(self, model, thinning, seed=seed)
        
        if model.solver.is_integer:
            raise TypeError("sampling does not work with integer problems :(")
        self.model = model.copy()

        self.thinning = thinning

        if nproj is None:
            self.nproj = int(min(len(self.model.variables)**3, 1e6))
        else:
            self.nproj = nproj
        self.n_samples = 0
        self.retries = 0
        
        # Careful, double underscore names are mangled to the class name
        self.problem = self._HRSampler__build_problem()

        # Set up a map from reaction -> forward/reverse variable
        var_idx = {v: idx for idx, v in enumerate(self.model.variables)}
        self.var_idx = np.array(
            [var_idx[v] for v in self.model.variables])
        self.warmup = None
        if seed is None:
            self._seed = int(time())
        else:
            self._seed = seed
        # Avoid overflow
        self._seed = self._seed % np.iinfo(np.int32).max

    def generate_fva_warmup(self):
        """
        Adapted from cobra.flux_analysis.sampling.py
        __________________________________________

        Generate the warmup points for the sampler.

        Generates warmup points by setting each flux as the sole objective
        and minimizing/maximizing it. Also caches the projection of the
        warmup points into the nullspace for non-homogeneous problems (only
        if necessary).
        """
        self.n_warmup = 0
        idx = np.hstack([self.var_idx])
        self.warmup = np.zeros((len(idx), len(self.model.variables)))
        self.model.objective = S.Zero
        self.model.objective.direction = "max"
        variables = self.model.variables
        for i in idx:
            # Omit fixed reactions
            if self.problem.variable_fixed[i]:
                self.model.logger.info("skipping fixed variable %s" % variables[i].name)
                continue
            self.model.objective.set_linear_coefficients({variables[i]: 1})
            self.model.slim_optimize()
            if not self.model.solver.status == OPTIMAL:
                self.model.logger.info(
                    "can not maximize variable %s, skipping it" % variables[
                        i].name)
                continue
            primals = self.model.solver.primal_values
            sol = [primals[v.name] for v in self.model.variables]
            self.warmup[self.n_warmup,] = sol
            self.n_warmup += 1
            # revert objective
            self.model.objective.set_linear_coefficients({variables[i]: 0})
        # Shrink warmup points to measure
        self.warmup = shared_np_array((self.n_warmup, len(variables)),
                                      self.warmup[0:self.n_warmup, ])


# Next, we redefine the analysis class as both inheriting from the
# GeneralizedHRSampler, and then the original classes. This will overwrite the
# inherited methods from HRSampler, while keeping the lower level ones from the
# samplers

class GeneralizedACHRSampler(GeneralizedHRSampler,ACHRSampler):
    def __init__(self, model, thinning=100, seed=None):
        """
        Adapted from cobra.flux_analysis.analysis
        __________________________________________
        Initialize a new ACHRSampler."""
        GeneralizedHRSampler.__init__(self, model, thinning, seed=seed)
        self.generate_fva_warmup()
        self.prev = self.center = self.warmup.mean(axis=0)
        np.random.seed(self._seed)

class GeneralizedOptGPSampler(GeneralizedHRSampler, OptGPSampler):
    def __init__(self, model, processes, thinning=100, seed=None):
        """
        Adapted from cobra.flux_analysis.sampling.py
        __________________________________________
        Initialize a new OptGPSampler."""
        GeneralizedHRSampler.__init__(self, model, thinning, seed=seed)
        self.generate_fva_warmup()
        self.processes = processes

        # This maps our saved center into shared memory,
        # meaning they are synchronized across processes
        self.center = shared_np_array((len(self.model.variables), ),
                                      self.warmup.mean(axis=0))


def sample(model, n, method="optgp", thinning=100, processes=1, seed=None):
    """
    Sample valid flux distributions from a thermo cobra_model.

    Function adapted from cobra.flux_analysis.sample to display all solver
    variables

    **Documentation adapted from cobra.flux_analysis.sample**

    1. 'optgp' (default) which uses the OptGPSampler that supports parallel
        analysis [1]_. Requires large numbers of samples to be performant
        (n < 1000). For smaller samples 'achr' might be better suited.

    or

    2. 'achr' which uses artificial centering hit-and-run. This is a single
       process method with good convergence [2]_.

    Parameters
    ----------
    model : pytfa.core.ThermoModel
        The cobra_model from which to sample variables.
    n : int
        The number of samples to obtain. When using 'optgp' this must be a
        multiple of `processes`, otherwise a larger number of samples will be
        returned.
    method : str, optional
        The analysis algorithm to use.
    thinning : int, optional
        The thinning factor of the generated analysis chain. A thinning of 10
        means samples are returned every 10 steps. Defaults to 100 which in
        benchmarks gives approximately uncorrelated samples. If set to one
        will return all iterates.
    processes : int, optional
        Only used for 'optgp'. The number of processes used to generate
        samples.
    seed : positive integer, optional
        The random number seed to be used. Initialized to current time stamp
        if None.

    Returns
    -------
    pandas.DataFrame
        The generated flux samples. Each row corresponds to a sample of the
        fluxes and the columns are the reactions.

    Notes
    -----
    The samplers have a correction method to ensure equality feasibility for
    long-running chains, however this will only work for homogeneous models,
    meaning models with no non-zero fixed variables or constraints (
    right-hand side of the equalities are zero).

    References
    ----------
    .. [1] Megchelenbrink W, Huynen M, Marchiori E (2014)
       optGpSampler: An Improved Tool for Uniformly Sampling the Solution-Space
       of Genome-Scale Metabolic Networks.
       PLoS ONE 9(2): e86587.
    .. [2] Direction Choice for Accelerated Convergence in Hit-and-Run Sampling
       David E. Kaufman Robert L. Smith
       Operations Research 199846:1 , 84-95
    """
    if method == "optgp":
        sampler = GeneralizedOptGPSampler(model, processes, thinning=thinning, seed=seed)
    elif method == "achr":
        sampler = GeneralizedACHRSampler(model, thinning=thinning, seed=seed)
    else:
        raise ValueError("method must be 'optgp' or 'achr'!")

    return sampler.sample(n, fluxes = False)



# -*- coding: utf-8 -*-
"""Standard constants definitions
"""

from math import log

TEMPERATURE_0 = 298.15  # K
MIN_PH = 3
MAX_PH = 9

# Used to calculate the transformed Gibbs energy of formation of specie with
# given pH and ionic strength using formula given by Goldberg and Tewari, 1991
# equation 4.4-10 in Alberty's book
DEBYE_HUCKEL_B_0 = 1.6
DEBYE_HUCKEL_A = 1.17582 / log(10)

A_LOT = 5000
A_LITTLE = 0.5
A_COUPLE = 2.5 # A couple is usually considered to be 2 or 3
MANY = 100


#!/usr/bin/env python

# -*- coding: utf-8 -*-

"""
.. module:: redgem
   :platform: Unix, Windows
   :synopsis: RedGEM Algorithm

.. moduleauthor:: pyTFA team

Model class
"""

from pytfa.redgem.network_expansion import NetworkExpansion
from pytfa.redgem.lumpgem import LumpGEM
from cobra import Reaction
from .utils import remove_blocked_reactions, set_medium
import yaml

class RedGEM():
    def __init__(self, gem, parameters_path, inplace=False):

        self.read_parameters(parameters_path)

        # If inplace is True, no deepcopy is performed : the modifications are applied directly onto the gem
        prepared_gem = set_medium(gem, self.params['medium'], inplace)

        self._gem = prepared_gem
        # This one is used to perform the lumping
        self._source_gem = prepared_gem.copy()

        self.logger = self._gem.logger

        self.fill_default_params()

        self.set_solver()

    def read_parameters(self, parameters_path):
        with open(parameters_path, 'r') as stream:
            try:
                self.params = yaml.safe_load(stream)
                print("Opened parameters file")
            except yaml.YAMLError as exc:
                print(exc)

    def fill_default_params(self):
        # If auto is activated, automatically extracts inorganics from the gem
        if "inorganic" not in self.params or self.params["inorganics"] == "auto":
            self.logger.info("Automatically computing inorganics to use")
            self.params["inorganics"] = self._extract_inorganics()
        if "growth_rate" not in self.params or self.params["growth_rate"] == "auto":
            self.logger.info("Setting minimal growth rate to 95% of the TFA solution")
            obj_val = self._source_gem.slim_optimize()
            self.logger.info("Setting minimal growth rate to {}".format(obj_val))
            self.params["growth_rate"] = 0.95*obj_val
        if "force_solve" not in self.params:
            self.params["force_solve"] = False
        if "timeout" not in self.params:
            self.logger.info("Using default timeout : 3600s")
            self.params["timeout"] = 3600
        if "feasibility" not in self.params:
            self.logger.info("Using default solver feasibility : 1e-9")
            self.params["feasibility"] = 1e-9
        else:
            # numbers like 1e-9 are detected as strings by yaml module
            # to enable their use, we cast them into floats
            try:
                self.params["feasibility"] = float(self.params["feasibility"])
            except ValueError as v:
                self.logger.error(v)

    def set_solver(self):
        if "solver" not in self.params or self.params["solver"].lower() == "auto":
            return None
        elif 'gurobi' in self.params["solver"].lower():
            solver = 'gurobi'
        elif 'cplex' in self.params["solver"].lower():
            solver = 'cplex'
        elif 'glpk' in self.params["solver"].lower():
            solver = 'glpk'
        else:
            solver = self.params["solver"]

        self._gem.solver = solver
        self._source_gem.solver = solver


    def run(self):
        # Extracting parameters
        core_subsystems = self.params["core_subsystems"]
        extracellular_system = self.params["extracellular_system"]
        biomass_rxn_ids = self.params["biomass_rxns"]

        biomass_rxns = [self._gem.reactions.get_by_id(x) for x in biomass_rxn_ids]
        main_bio_rxn = biomass_rxns[0]

        growth_rate = self.params["growth_rate"]

        small_metabolites = self.params["small_metabolites"]
        cofactor_pairs = self.params["cofactor_pairs"]
        # Flatten cofactor_pairs list
        cofactors = [cofactor for pair in cofactor_pairs for cofactor in pair]
        inorganics = self.params["inorganics"]

        d = self.params["d"]
        n = self.params["n"]
        lump_method = self.params["lump_method"]

        force_solve = self.params["force_solve"]
        timeout = self.params["timeout"]
        self._gem.solver.configuration.tolerances.feasibility = self.params["feasibility"]
        self._gem.solver.configuration.tolerances.integrality = self.params["feasibility"]
        self._source_gem.solver.configuration.tolerances.feasibility = self.params["feasibility"]
        self._source_gem.solver.configuration.tolerances.integrality = self.params["feasibility"]

        self.logger.info("Computing network expansion...")
        expander = NetworkExpansion(self._gem, core_subsystems, extracellular_system,
                                    cofactors, small_metabolites, inorganics,
                                    d, n)
        reduced_gem = expander.run()
        self.logger.info("Done.")

        # Add the expansion to core reactions
        core_reactions = reduced_gem.reactions

        self.logger.info("Computing lumps...")
        lumper = LumpGEM(self._source_gem, core_reactions, self.params)
        lumps = lumper.compute_lumps(force_solve, method = lump_method)
        self.logger.info("Done.")

        self.logger.info("Create final network...")
        to_add = [x for x in biomass_rxns
                            +lumper._exchanges
                            +lumper._transports
                            +lumper._rcore
                  if not x.id in reduced_gem.reactions]
        reduced_gem.add_reactions(to_add)

        for rxns in lumps.values():
            the_lumps = [add_lump(reduced_gem,rxn,id_suffix='_{}'.format(e))
                         for e,rxn in enumerate(rxns)]
            # reduced_gem.add_reactions(rxns)
        self.logger.info("Done.")

        reduced_gem.objective = main_bio_rxn
        reduced_gem.reactions.get_by_id(main_bio_rxn.id).lower_bound = growth_rate

        if self.params['remove_blocked_reactions']:
            self.logger.info('Detecting blocked reactions')
            # Remove blocked reactions
            nrxn_1 = len(reduced_gem.reactions)
            self.removed_reactions = remove_blocked_reactions(reduced_gem)
            nrxn_2 = len(reduced_gem.reactions)
            self.logger.info('Removed {} blocked reaction with '
                             'FVA post-processing'.format(nrxn_1-nrxn_2))

        if main_bio_rxn.id not in reduced_gem.reactions:
            raise RuntimeError('Main Biomass reaction appears blocked')


        # For debugging purposes
        self.lumper = lumper
        main_bio_rxn.lower_bound = 0

        return reduced_gem

    def _extract_inorganics(self):
        """
        Extract inorganics from self._gem based on their formula

        :return: list of inorganics metabolites
        """

        inorganics = []
        for met in self._gem.metabolites:
            if not met.elements == {}: # Edge case
                # met is inorganic if it has 0 carbon in its formula
                if (not 'C' in met.elements) or met.elements['C'] <= 0:
                    inorganics.append(met.id)

        return inorganics

def add_lump(model, lump_object, id_suffix=''):
    new = Reaction(id = lump_object.id_+id_suffix)
    model.add_reaction(new)
    new.add_metabolites(lump_object.metabolites)
    new.gene_reaction_rule = lump_object.gene_reaction_rule
    new.subnetwork = lump_object.subnetwork

    return new



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Pre-tuned configurations for faster solving

"""

def dg_relax_config(model):
    """

    :param model:
    :return:
    """
    # grbtune output on a hard model :
    #
    # Tested 6992 parameter sets in 46793.78s
    #
    # Baseline parameter set: mean runtime 142.09s
    #
    # Improved parameter set 1 (mean runtime 3.27s):
    #
    # 	NormAdjust 0
    # 	BranchDir 1
    # 	DegenMoves 0
    # 	Heuristics 0
    # 	MIPFocus 1
    # 	Cuts 3
    #
    # Improved parameter set 2 (mean runtime 3.30s):
    #
    # 	NormAdjust 0
    # 	BranchDir 1
    # 	DegenMoves 0
    # 	Heuristics 0.001
    # 	PreSparsify 0
    #
    # Improved parameter set 3 (mean runtime 3.34s):
    #
    # 	NormAdjust 0
    # 	BranchDir 1
    # 	DegenMoves 0
    # 	Heuristics 0.001
    #
    # Improved parameter set 4 (mean runtime 5.22s):
    #
    # 	NormAdjust 1
    # 	BranchDir 1
    # 	DegenMoves 0
    #
    # Improved parameter set 5 (mean runtime 7.18s):
    #
    # 	BranchDir 1
    # 	DegenMoves 0

    if model.solver.interface.__name__ == 'optlang.gurobi_interface':
        model.solver.problem.Params.NormAdjust = 0
        model.solver.problem.Params.BranchDir = 1
        model.solver.problem.Params.DegenMoves = 0
        model.solver.problem.Params.Heuristics = 0.001
        model.solver.problem.Params.Cuts = 3
        model.solver.problem.Params.Presolve = 2
        model.solver.problem.Params.Method = 0




# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Variable declarations

"""
from abc import ABCMeta

from distutils.version import LooseVersion
from optlang import __version__ as OPTLANG_VER

from ..utils.str import camel2underscores
from .meta import ABCRequirePrefixMeta

from warnings import warn


op_replace_dict = {
    ' + ': '_ADD_',
    ' - ': '_SUB_',
    ' * ': '_MUL_',
    ' / ': '_DIV_',
}

###################################################
###                  VARIABLES                  ###
###################################################

class GenericVariable(metaclass=ABCRequirePrefixMeta):
    """
    Class to represent a generic variable. The purpose is that the interface
    is instantiated on initialization, to follow the type of interface used
    by the problem, and avoid incompatibilities in optlang

    Attributes:

        :id: Used for DictList comprehension. Usually points back at a
        enzyme or reaction id for ease of linking. Should be unique given
        a variable type.
        :name: Should be a concatenation of the id and a prefix that is
        specific to the variable type. will be used to address the variable at
        the solver level, and hence should be unique in the whole cobra_model
        :cobra_model: the cobra_model hook.
        :variable: links directly to the cobra_model representation of tbe variable
    """
    prefix = NotImplemented

    @property
    def __attrname__(self):
        """
        Name the attribute the instances will have
        Example: GenericVariable -> generic_variable
        :return:
        """
        return camel2underscores(self.__class__.__name__)

    def __init__(self, id_='', model=None, hook=None, queue=False, scaling_factor=1,
                 **kwargs):
        """

        :param id_: will be used to identify the variable
            (name will be a concat of this and a prefix)
        :param model: the cobra.Model object
        :param queue: whether or not to queue the variable for update object
        :param scaling_factor: scaling factor used in self.scaled, useful for
                                adimensionalisation of constraints
        :param kwargs: stuff you want to pass to the variable constructor
        """
        self.hook = hook
        self._id = id_
        self._model = model
        self.kwargs = kwargs
        self._name = self.make_name()
        self.get_interface(queue)
        self.prefix = ''
        self._scaling_factor = scaling_factor

    def get_interface(self, queue):
        """
        Called upon completion of __init__, initializes the value of self.var,
        which is returned upon call, and stores the actual interfaced variable.

        :return: instance of Variable from the problem
        """

        if not self.name in self.model.variables:
            variable = self.model.problem.Variable(name = self.name, **self.kwargs)
            if not queue:
                self.model.add_cons_vars(variable, sloppy=self.model.sloppy)
            else:
                self.model._var_queue.append(variable)
        else:
            self.variable = self.model.variables.get(self.name)

    def make_name(self):
        """
        Needs to be overridden by the subclass, concats the id with a
         prefix

        :return: None
        """
        return self.prefix + self.id

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def id(self):
        """
        for cobra.thermo.DictList compatibility
        :return:
        """
        return self._id

    @property
    def variable(self):
        return self.model.variables[self.name]

    @variable.setter
    def variable(self,value):
        self.model.variables[self.name] = value

    @property
    def scaling_factor(self):
        return self._scaling_factor

    @property
    def unscaled(self):
        """
        If the scaling factor of quantity X is a, it is represented by
        the variable X_hat = X/a.
        This returns X = a.X_hat
        Useful for nondimensionalisation of variables and constraints

        :return: The variable divided by its scaling factor
        """
        return self.scaling_factor * self

    @property
    def value(self):
        try:
            return self.model.solution.raw[self.name]
        except AttributeError:
            warn('Model need to be optimized to get a value for this variable')

    @property
    def unscaled_value(self):
        try:
            return self.scaling_factor * self.value
        except AttributeError:
            warn('Model need to be optimized to get a value for this variable')

    @property
    def model(self):
        return self._model
    #
    # @variable.setter
    # def variable(self, value):
    #     self._variable = value

    @property
    def type(self):
        return self.variable.type

    def test_consistency(self, other):
        """
        Tests whether a candidate to an operation is of the right type and is
        from the same problem

        :param other: an object
        :return: None
        """

        assert (isinstance(other, GenericVariable))
        # assert (other.problem == self.problem)

    def get_operand(self, other):
        """
        For operations, choose if the operand is a GenericVariable, in which
        we return its optlang variable, or something else (presumably a numeric)
        and we let optlang decide what to do

        :param other:
        :return:
        """
        if isinstance(other, GenericVariable):
            return other.variable
        else:  # Let optlang decide what to do
            return other

    #########################################################################
    # We redefine all the following operations to return the equivalent     #
    # expression one would get by applying the said operation to the        #
    # self.variable attribute                                               #
    #########################################################################

    def __add__(self, other):
        """
        Adding either two variables together or a variable and a numeric
        results in a new variable
        :param other:
        :return: a new Generic Variable
        """
        operand = self.get_operand(other)

        new_variable = self.variable + operand

        return self.make_result(new_variable)

    def __radd__(self, other):
        """
        Take priority on symmetric arithmetic operation
        :param other:
        :return:
        """

        return self.__add__(other)

    def __sub__(self, other):
        """
        Substracting either two variables together or a variable and a numeric
        results in a new variable
        :param other:
        :return: a new Generic Variable
        """
        operand = self.get_operand(other)

        new_variable = self.variable - operand

        return self.make_result(new_variable)

    def __rsub__(self, other):
        """
        Take priority on symmetric arithmetic operation
        :param other:
        :return:
        """

        operand = self.get_operand(other)

        new_variable = operand - self.variable

        return self.make_result(new_variable)

    def __mul__(self, other):
        """
        Multiplying either two variables together or a variable and a numeric
        results in a new variable
        :param other:
        :return: a new Generic Variable
        """
        operand = self.get_operand(other)

        new_variable = self.variable * operand

        return self.make_result(new_variable)

    def __rmul__(self, other):
        """
        Take priority on symmetric arithmetic operation
        :param other:
        :return:
        """
        return self.__mul__(other)

    def __truediv__(self, other):
        """
        Dividing either two variables together or a variable and a numeric
        results in a new variable
        :param other:
        :return: a new Generic Variable
        """
        operand = self.get_operand(other)

        new_variable = self.variable / operand

        return self.make_result(new_variable)

    def __rtruediv__(self, other):
        """
        Take priority on symmetric arithmetic operation
        :param other:
        :return:
        """

        operand = self.get_operand(other)

        new_variable = operand / self.variable

        return self.make_result(new_variable)


    def make_result(self, new_variable):
        """
        Returns a Sympy expression
        :param new_variable:
        :return:
        """

        # if isinstance(new_variable, self.cobra_model.problem.Variable):
        #     new_name = new_variable.name
        # else: #It is an expression
        #     new_name  = new_variable.__str__()
        #     for k,v in op_replace_dict.items():
        #         new_name = new_name.replace(k,v)

        # result = GenericVariable(new_name, self.cobra_model)
        # result.variable = new_variable

        # return result
        return new_variable

    def __repr__(self):
        return self.variable.__repr__()

def get_binary_type():
    """
    FIX : We enforce type to be integer instead of binary, else optlang does
    not allow to set the binary variable bounds to anything other than (0,1)
    You might want to set it at (0,0) to enforce directionality for example
    """
    if LooseVersion(OPTLANG_VER) < LooseVersion('1.2.3'):
        return 'integer'
    else:
        return 'binary'

class ModelVariable(GenericVariable):
    """
    Class to represent a variable attached to the model
    """

    def __init__(self, model, id_, **kwargs):
        if not 'lb' in kwargs:
            kwargs['lb'] = 0
        GenericVariable.__init__(self,
                                 id_= id_,
                                 model=model,
                                 hook=model,
                                 **kwargs)
    prefix = 'MODV_'

class GeneVariable(GenericVariable):
    """
    Class to represent a gene variable
    """

    prefix = 'GV_'

    def __init__(self, gene, **kwargs):
        model = gene.model


        if not 'lb' in kwargs:
            kwargs['lb'] = 0
        if not 'ub' in kwargs:
            pass

        GenericVariable.__init__(self, model=model, hook=gene,
                                 **kwargs)

    @property
    def gene(self):
        return self.hook

    @property
    def id(self):
        return self.gene.id

    @property
    def model(self):
        return self.gene.model

class BinaryVariable(GenericVariable):
    """
    Class to represent a generic binary variable
    """

    def __init__(self, id_, model, **kwargs):

        if not 'lb' in kwargs:
            kwargs['lb'] = 0
        if not 'ub' in kwargs:
            kwargs['ub'] = 1

        GenericVariable.__init__(self,
                                 id_,
                                 model = model,
                                 type=get_binary_type(),
                                 **kwargs)

    prefix = 'B_'

class ReactionVariable(GenericVariable):
    """
    Class to represent a variable attached to a reaction
    """

    def __init__(self, reaction, **kwargs):
        model = reaction.model

        GenericVariable.__init__(self,
                                 model=model,
                                 hook=reaction,
                                 **kwargs)

    @property
    def reaction(self):
        return self.hook

    @property
    def id(self):
        return self.reaction.id

    @property
    def model(self):
        return self.reaction.model

    prefix = 'RV_'

class MetaboliteVariable(GenericVariable):
    """
    Class to represent a variable attached to a enzyme
    """

    def __init__(self, metabolite, **kwargs):
        model = metabolite.model

        GenericVariable.__init__(self,
                                 model=model,
                                 hook=metabolite,
                                 **kwargs)

    @property
    def metabolite(self):
        return self.hook

    @property
    def id(self):
        return self.metabolite.id

    @property
    def model(self):
        return self.metabolite.model

    prefix = 'MV_'

class ForwardUseVariable(ReactionVariable, BinaryVariable):
    """
    Class to represent a forward use variable, a type of binary variable used to
    enforce forward directionality in reaction net fluxes
    """

    def __init__(self, reaction, **kwargs):
        ReactionVariable.__init__(self, reaction,
                                  type=get_binary_type(),
                                  **kwargs)

    prefix = 'FU_'

class BackwardUseVariable(ReactionVariable, BinaryVariable):
    """
    Class to represent a backward use variable, a type of binary variable used
    to enforce backward directionality in reaction net fluxes
    """

    def __init__(self, reaction, **kwargs):
        if not 'lb' in kwargs:
            kwargs['lb'] = 0
        if not 'ub' in kwargs:
            kwargs['ub'] = 1

        ReactionVariable.__init__(self, reaction,
                                  type=get_binary_type(),
                                  **kwargs)

    prefix = 'BU_'

class ForwardBackwardUseVariable(ReactionVariable, BinaryVariable):
    """
    Class to represent a type of binary variable used to tell whether the
    reaction is active or not such that:
        FU + BU + BFUSE = 1
    """

    def __init__(self, reaction, **kwargs):
        if not 'lb' in kwargs:
            kwargs['lb'] = 0
        if not 'ub' in kwargs:
            kwargs['ub'] = 1

        ReactionVariable.__init__(self, reaction,
                                  type=get_binary_type(),
                                  **kwargs)

    prefix = 'BFUSE_'

class LogConcentration(MetaboliteVariable):
    """
    Class to represent a log concentration of a enzyme
    """

    prefix = 'LC_'

class DeltaGErr(ReactionVariable):
    """
    Class to represent a DeltaGErr
    """

    prefix = 'DGE_'

class DeltaG(ReactionVariable):
    """
    Class to represent a DeltaG
    """

    prefix = 'DG_'

class DeltaGstd(ReactionVariable):
    """
    Class to represent a DeltaG^o (naught) - standard conditions
    """

    prefix = 'DGo_'

class ThermoDisplacement(ReactionVariable):
    """
    Class to represent the thermodynamic displacement of a reaction
    \Gamma = -\DeltaG/RT
    """

    prefix = 'LnGamma_'

class PosSlackVariable(ReactionVariable):
    """
    Class to represent a positive slack variable for relaxation problems
    """

    def __init__(self,reaction,**kwargs):
        ReactionVariable.__init__(self, reaction, **kwargs)

    prefix = 'PosSlack_'

class NegSlackVariable(ReactionVariable):
    """
    Class to represent a negative slack variable for relaxation problems
    """

    def __init__(self,reaction,**kwargs):
        ReactionVariable.__init__(self, reaction, **kwargs)

    prefix = 'NegSlack_'

class PosSlackLC(MetaboliteVariable):

    prefix = 'PosSlackLC_'

class NegSlackLC(MetaboliteVariable):

    prefix = 'NegSlackLC_'

class LinearizationVariable(ModelVariable):
    """
    Class to represent a product A*B when performing linearization of the
    model
    """
    prefix = 'LZ_'



"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Some tools around COBRApy models used by pyTFA


"""
import re

Formula_regex = re.compile("([A-Z][a-z]*)([0-9]*)")


def check_reaction_balance(reaction, proton = None):
    """
    Check the balance of a reaction, and eventually add protons to balance
    it

    :param cobra.thermo.reaction.Reaction reaction: The reaction to check the
        balance of.
    :param cobra.thermo.metabolite.Metabolite proton: *Optional* The proton to add
        to the reaction to balance it.

    :returns: The balance of the reaction :

        * ``drain flux``
        * ``missing structures``
        * ``balanced``
        * ``N protons added to reactants`` with ``N`` a :any:`float`
        * ``N protons added to products`` with ``N`` a :any:`float`
        * ``missing atoms``

    :rtype: str

    If ``proton`` is provided, this function will try to balance the equation
    with it, and return the result.

    If no ``proton`` is provided, this function will not try to balance the
    equation.

    .. warning::
       This function does not verify if ``proton`` is in the correct
       compartment, so make sure you provide the ``proton`` belonging to the
       correct compartment !

    """
    # Sum of the total charges of both sides (reactants counted negatively,
    # products counted positively)
    sum_charge = 0

    if len(reaction.metabolites) == 1:
        return 'drain flux'

    # Count the atoms on each side
    Atoms_sum = [0] * 26

    for metabolite in reaction.metabolites:
        if metabolite.formula == 'NA' or \
           metabolite.formula is None:
            return 'missing structures'
        metCharge = metabolite.thermo.charge_std
        metCoeff = reaction.metabolites[metabolite]

        sum_charge += metCharge * metCoeff

        # Parse the formula to extract atoms
        atoms = Formula_regex.findall(metabolite.formula)

        for atom in atoms:
            try:
                id_ = ['C', 'N', 'O', 'H', 'P', 'Na', 'Mg', 'S', 'Cl', 'K',
                'Ca', 'Mn', 'Fe', 'Ni', 'Co', 'Cu', 'Zn', 'As', 'Se', 'Ag',
                'Cd', 'W', 'Hg', 'R', 'Mo', 'X'].index(atom[0])
            except ValueError:
                # Jorge Carrasco: Can we use warnings here?
                reaction.model.logger.warning(
                    'Warning : ' + metabolite.formula + '/' + atom[0]
                )
                continue

            Atoms_sum[id_] += metCoeff * (int(atom[1]) if atom[1] else 1)

    # Find unbalanced atoms
    nonNull = [i for i in range(26) if Atoms_sum[i] != 0]

    # If everything is balanced
    if len(nonNull) == 0 and sum_charge == 0:
        return 'balanced'

    # Add protons to compensate if possible
    if (proton and len(nonNull) == 1
        and nonNull[0] == 3
        and Atoms_sum[3] == sum_charge):

        reaction.add_metabolites({proton: -sum_charge})

        if sum_charge > 0:
            return str(sum_charge) + ' protons added to reactants'
        else:
            return str(-sum_charge) + ' protons added to products'

    return 'missing atoms'


def find_transported_mets(reaction):
    """
    Get a list of the transported metabolites of the reaction.

    :param cobra.thermo.reaction.Reaction reaction: The reaction to get the
        transported metabolites of

    :returns: A dictionnary of the transported metabolites.
        The index corresponds to the seed_id of the transported enzyme

        The value is a dictionnairy with the following values:

            * coeff (:any:`float`):
                The stoechiomectric coefficient of the enzyme
            * reactant (:any:`cobra.thermo.enzyme.Metabolite`):
                The reactant of the reaction corresponding to the transported
                enzyme
            * product (:any:`cobra.thermo.enzyme.Metabolite`):
                The product of the reaction corresponding to the transported
                enzyme

    A transported enzyme is defined as a enzyme which is a product and a
    reactant of a reaction. We can distinguish them thanks to their seed_ids.

    """

    # First, we get all the reactants of the reaction
    reactants_coeffs = {}

    for met in reaction.reactants:
        reactants_coeffs[met.annotation['seed_id']] = {
            'coeff':reaction.metabolites[met],
            'met':met
        }

    # Initialize the dictionnary we will fill with the transported metabolites
    trans_coeffs = {}

    for met in reaction.products:
        seed_id = met.annotation['seed_id']

        # If the seed_id also corresponds to a reactant, we add it to the result
        if seed_id in reactants_coeffs:
            trans_coeffs[seed_id] = {
                'coeff'    : max(
                                abs(reactants_coeffs[seed_id]['coeff']),
                                abs(reaction.metabolites[met])
                            ),
                'reactant' : reactants_coeffs[seed_id]['met'],
                'product'  : met
            }

    return trans_coeffs


def check_transport_reaction(reaction):
    """

    Check if a reaction is a transport reaction

    :param cobra.thermo.reaction.Reaction reaction: The reaction to check

    :returns: Whether the reaction is a transport reaction or not
    :rtype: bool

    A transport reaction is defined as a reaction that has the same compound
    as a reactant and a product. We can distinguish them thanks to their seed_ids.
    If they have one
    If not, use met_ids and check if they are the same, minus compartment

    """
    seed_ids = {}
    try:
        for reactant in reaction.reactants:
            seed_ids[reactant.annotation['seed_id']] = True

        for product in reaction.products:
            if product.annotation['seed_id'] in seed_ids:
                return True
    except KeyError:
        reactants_ids = [x.id.replace(x.compartment,'') for x in reaction.reactants]
        product_ids = [x.id.replace(x.compartment,'') for x in reaction.products]

        return set(reactants_ids) == set(product_ids)

    return False


def is_same_stoichiometry(this_reaction, that_reaction):
    this_met_dict = {k.id:v for k,v in this_reaction.metabolites.items()}
    that_met_dict = {k.id:v for k,v in that_reaction.metabolites.items()}
    return this_met_dict == that_met_dict


def is_exchange(rxn):
    return len(rxn.metabolites) == 1


def get_reaction_compartment(reaction):
    """Get the compartment of a reaction to then prepare it for conversion."""
    comp = None
    for met in reaction.metabolites:
        if comp is None:
            comp = met.compartment
        elif met.compartment != comp:
            comp = 'c'
    return comp



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Relaxation of models with constraint too tight

"""

from collections import OrderedDict
from copy import deepcopy

from tqdm import tqdm
import pandas as pd
from cobra.util.solver import set_objective
from optlang.exceptions import SolverError

from .constraints import NegativeDeltaG
from .config import dg_relax_config
from .utils import get_solution_value_for_variables, chunk_sum, symbol_sum
from .variables import PosSlackVariable, NegSlackVariable, DeltaGstd, \
    LogConcentration, NegSlackLC, PosSlackLC
from ..utils import numerics

BIGM = numerics.BIGM
BIGM_THERMO = numerics.BIGM_THERMO
BIGM_DG = numerics.BIGM_DG
BIGM_P = numerics.BIGM_P
EPSILON = numerics.EPSILON

def relax_dgo_gurobi(model, relax_obj_type = 0):

    the_cons = [x.constraint._internal_constraint
                for x in model.get_constraints_of_type(NegativeDeltaG)]
    cons_penalities = [1]*len(the_cons)

    # the_vars = [x._internal_variable for x in model.variables]
    # vars_penalities = [1]*len(the_vars)

    grm = model.solver.problem.feasRelax(relaxobjtype=relax_obj_type,
                                         minrelax=True,
                                         constrs=the_cons,
                                         # vars=the_vars,
                                         # lbpen=vars_penalities,
                                         # ubpen=vars_penalities,
                                         vars=None,
                                         lbpen=None,
                                         ubpen=None,
                                         rhspen=cons_penalities)

    return grm

def relax_dgo(tmodel, reactions_to_ignore=(), solver=None, in_place = False):
    """
    :param t_tmodel:
    :type t_tmodel: pytfa.thermo.ThermoModel:
    :param reactions_to_ignore: Iterable of reactions that should not be relaxed
    :param solver: solver to use (e.g. 'optlang-glpk', 'optlang-cplex',
        'optlang-gurobi'
    :return: a cobra_model with relaxed bounds on standard Gibbs free energy
    """

    if solver is None:
        solver = tmodel.solver.interface

    # Create a copy of the cobra_model on which we will perform the slack addition
    slack_model = deepcopy(tmodel)
    slack_model.solver = solver
    slack_model.name = 'SlackModel '+tmodel.name
    slack_model.id = 'SlackModel_'+tmodel.id

    # Ensure the lazy updates are all done
    slack_model.repair()

    if not in_place:
        # Create a copy that will receive the relaxation
        relaxed_model = deepcopy(tmodel)
        relaxed_model.solver = solver
        relaxed_model.name = 'RelaxedModel '+tmodel.name
        relaxed_model.id = 'RelaxedModel_'+tmodel.id
        relaxed_model.repair()
    else:
        relaxed_model = slack_model

    dg_relax_config(slack_model)

    original_objective = relaxed_model.objective


    # Do not relax if cobra_model is already optimal
    # try:
    #     solution = tmodel.optimize()
    # except SolverError as SE:
    #     status = tmodel.solver.status
    #     tmodel.logger.error(SE)
    #     tmodel.logger.warning('Solver status: {}'.format(status))
    # if tmodel.solver.status == OPTIMAL:
    #     raise Exception('Model is already optimal')

    # Find variables that represent standard Gibbs Energy change
    my_dgo = relaxed_model.get_variables_of_type(DeltaGstd)

    # Find constraints that represent negativity of Gibbs Energy change
    my_neg_dg = slack_model.get_constraints_of_type(NegativeDeltaG)

    changes = OrderedDict()
    objective_symbols = []

    slack_model.logger.info('Adding slack constraints')

    slack_model.solver.update()

    for this_neg_dg in tqdm(my_neg_dg, desc='adding slacks'):

        # If there is no thermo, or relaxation forbidden, pass
        if this_neg_dg.id in reactions_to_ignore or this_neg_dg.id not in my_dgo:
            continue

        # Create the negative and positive slack variables
        # We can't queue them because they will be in an expression to declare
        # the constraint
        neg_slack = slack_model.add_variable(NegSlackVariable,
                                             this_neg_dg.reaction, lb=0,
                                             ub=BIGM_DG,
                                             queue=False)
        pos_slack = slack_model.add_variable(PosSlackVariable,
                                             this_neg_dg.reaction, lb=0,
                                             ub=BIGM_DG,
                                             queue=False)

        # Create the new constraint by adding the slack variables to the
        # negative delta G constraint (from the initial cobra_model)
        new_expr = this_neg_dg.constraint.expression
        new_expr += (pos_slack - neg_slack)

        this_reaction = this_neg_dg.reaction
        # # Remove former constraint to override it
        # slack_model.remove_constraint(slack_model._cons_dict[this_neg_dg.name])
        #
        # # Add the new variant
        # slack_model.add_constraint(NegativeDeltaG,
        #                            this_reaction,
        #                            expr=new_expr,
        #                            lb=0,
        #                            ub=0,
        #                            queue=True)

        this_neg_dg.change_expr(new_expr)

        # Update the objective with the new variables
        objective_symbols += [neg_slack,  pos_slack]

    # objective = chunk_sum(objective_symbols)
    objective = symbol_sum(objective_symbols)

    # Change the objective to minimize slack
    set_objective(slack_model, objective)

    # Update variables and constraints references
    slack_model.repair()

    slack_model.logger.info('Optimizing slack model')
    # Relax
    slack_model.objective.direction = 'min'
    relaxation = slack_model.optimize()

    # Extract the relaxation values from the solution, by type
    relaxed_model.logger.info('Extracting relaxation')
    my_neg_slacks = slack_model.get_variables_of_type(NegSlackVariable)
    my_pos_slacks = slack_model.get_variables_of_type(PosSlackVariable)

    neg_slack_values = get_solution_value_for_variables(relaxation,
                                                        my_neg_slacks)
    pos_slack_values = get_solution_value_for_variables(relaxation,
                                                        my_pos_slacks)

    epsilon = relaxed_model.solver.configuration.tolerances.feasibility
    relaxed_model.repair()
    relaxed_model.solver.update()
    # Apply reaction delta G standard bound change
    for this_reaction in tqdm(relaxed_model.reactions, desc = 'applying slack'):
        # No thermo, or relaxation forbidden
        if this_reaction.id in reactions_to_ignore or this_reaction.id not in my_dgo:
            continue

        # Get the standard delta G variable
        the_dgo = my_dgo.get_by_id(this_reaction.id)

        # Get the relaxation
        dgo_delta_lb = \
            neg_slack_values[my_neg_slacks \
                .get_by_id(this_reaction.id).name]
        dgo_delta_ub = \
            pos_slack_values[my_pos_slacks \
                .get_by_id(this_reaction.id).name]

        if in_place:
            the_neg_slack = my_neg_slacks.get_by_id(this_reaction.id)
            the_neg_slack_value = slack_model.solution.raw[the_neg_slack.name]
            the_neg_slack.variable.lb = the_neg_slack_value - epsilon
            the_neg_slack.variable.ub = the_neg_slack_value + epsilon

            the_pos_slack = my_pos_slacks.get_by_id(this_reaction.id)
            the_pos_slack_value = slack_model.solution.raw[the_pos_slack.name]
            the_pos_slack.variable.lb = the_pos_slack_value - epsilon
            the_pos_slack.variable.ub = the_pos_slack_value + epsilon

        # Apply reaction delta G standard bound change
        if dgo_delta_lb > 0 or dgo_delta_ub > 0:

            # Store previous values
            previous_dgo_lb = the_dgo.variable.lb
            previous_dgo_ub = the_dgo.variable.ub

            if not in_place:
                # Apply change
                the_dgo.variable.lb -= (dgo_delta_lb + epsilon)
                the_dgo.variable.ub += (dgo_delta_ub + epsilon)

            # If needed, store that in a report table
            changes[this_reaction.id] = [
                previous_dgo_lb,
                previous_dgo_ub,
                dgo_delta_lb,
                dgo_delta_ub,
                the_dgo.variable.lb,
                the_dgo.variable.ub]


    relaxed_model.repair()
    relaxed_model.logger.info('Testing relaxation')

    relaxed_model.objective = original_objective
    relaxed_model.objective.direction = 'max'

    relaxed_model.optimize()

    if len(changes) == 0:
        # The model is infeasible or something went wrong
        tmodel.logger.error('Relaxation could not complete '
                            '(no DeltaG relaxation found)')
        return relaxed_model, slack_model, None

    # Format relaxation
    relax_table = pd.DataFrame.from_dict(changes,
                                         orient = 'index')
    relax_table.columns = [ 'lb_in',
                            'ub_in',
                            'lb_change',
                            'ub_change',
                            'lb_out',
                            'ub_out']

    return relaxed_model, slack_model, relax_table



def relax_lc(tmodel, metabolites_to_ignore = (), solver = None):
    """

    :param metabolites_to_ignore:
    :param in_tmodel:
    :type in_tmodel: pytfa.thermo.ThermoModel:
    :param min_objective_value:
    :return:
    """

    if solver is None:
        solver = tmodel.solver

    # Create a copy of the cobra_model on which we will perform the slack addition
    slack_model = deepcopy(tmodel)
    slack_model.solver = solver

    # Create a copy that will receive the relaxation
    relaxed_model = deepcopy(tmodel)
    relaxed_model.solver = solver

    # Do not relax if cobra_model is already optimal
    try:
        solution = tmodel.optimize()
    except SolverError as SE:
        status = tmodel.solver.status
        tmodel.logger.error(SE)
        tmodel.logger.warning('Solver status: {}'.format(status))
    # if tmodel.solver.status == OPTIMAL:
    #     raise Exception('Model is already optimal')

    # Find variables that represent standard Gibbs Energy change
    my_lc = relaxed_model.get_variables_of_type(LogConcentration)

    # Find constraints that represent negativity of Gibbs Energy change
    my_neg_dg = slack_model.get_constraints_of_type(NegativeDeltaG)

    changes = OrderedDict()
    objective = 0

    pos_slack = dict()
    neg_slack = dict()

    for this_lc in my_lc:
        if this_lc.name in metabolites_to_ignore:
            continue

        neg_slack[this_lc.name] = slack_model.add_variable(NegSlackLC,
                                             this_lc,
                                             lb= 0,
                                             ub= BIGM_DG)

        pos_slack[this_lc.name] = slack_model.add_variable(PosSlackLC,
                                             this_lc,
                                             lb= 0,
                                             ub= BIGM_DG)

        # Update the objective with the new variables
        objective += (neg_slack[this_lc.name] + pos_slack[this_lc.name])

    for this_neg_dg in my_neg_dg:

        # If there is no thermo, or relaxation forbidden, pass
        if this_neg_dg.id not in my_neg_dg:
            continue


        subs_dict = {k: slack_model.variables.get(k.name) \
                     for k in this_neg_dg.constraint.variables}

        # Create the new constraint by adding the slack variables to the
        # negative delta G constraint (from the initial cobra_model)
        new_expr = this_neg_dg.constraint.expression.subs(subs_dict)

        for this_var in this_neg_dg.constraint.variables:
            if not this_var.name in neg_slack:
                continue

            met_id = pos_slack[this_var.name].id
            the_met = slack_model.metabolites.get_by_id(met_id)
            stoich = this_neg_dg.reaction.metabolites[the_met]
            new_expr += slack_model.RT * stoich \
                        * \
            (pos_slack[this_var.name] - neg_slack[this_var.name])

        # Remove former constraint to override it
        slack_model.remove_constraint(this_neg_dg)

        # Add the new variant
        slack_model.add_constraint(NegativeDeltaG, this_neg_dg.reaction,
                                   expr=new_expr, lb=0, ub=0)

    # Change the objective to minimize slack
    set_objective(slack_model, objective)

    # Update variables and constraints references
    slack_model.repair()

    # Relax
    slack_model.objective.direction = 'min'
    relaxation = slack_model.optimize()

    # Extract the relaxation values from the solution, by type
    my_neg_slacks = slack_model.get_variables_of_type(NegSlackLC)
    my_pos_slacks = slack_model.get_variables_of_type(PosSlackLC)

    neg_slack_values = get_solution_value_for_variables(relaxation,
                                                        my_neg_slacks)
    pos_slack_values = get_solution_value_for_variables(relaxation,
                                                        my_pos_slacks)
    epsilon = relaxed_model.solver.configuration.tolerances.feasibility

    for this_met in relaxed_model.metabolites:
        # No thermo, or relaxation forbidden
        if this_met.id in metabolites_to_ignore or this_met.id not in my_lc:
            continue

        # Get the standard delta G variable
        the_lc = my_lc.get_by_id(this_met.id)

        # Get the relaxation
        lc_delta_lb = neg_slack_values[
            my_neg_slacks \
            .get_by_id(this_met.id).name]
        lc_delta_ub = \
            pos_slack_values[my_pos_slacks \
            .get_by_id(this_met.id).name]

        # Apply reaction delta G standard bound change
        if lc_delta_lb > 0 or lc_delta_ub > 0:

            # Store previous values
            previous_lc_lb = the_lc.variable.lb
            previous_lc_ub = the_lc.variable.ub

            # Apply change
            the_lc.variable.lb -= (lc_delta_lb + epsilon)
            the_lc.variable.ub += (lc_delta_ub + epsilon)

            # If needed, store that in a report table
            changes[this_met.id] = [
                previous_lc_lb,
                previous_lc_ub,
                lc_delta_lb,
                lc_delta_ub,
                the_lc.variable.lb,
                the_lc.variable.ub]

    # Obtain relaxation
    relaxed_model.optimize()

    # Format relaxation
    relax_table = pd.DataFrame.from_dict(changes,
                                         orient = 'index')
    relax_table.columns = [ 'lb_in',
                            'ub_in',
                            'lb_change',
                            'ub_change',
                            'lb_out',
                            'ub_out']

    tmodel.logger.info('\n' + relax_table.__str__())

    relaxed_model.relaxation = relax_table

    return relaxed_model, slack_model, relax_table



"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Some tools used by pyTFA


"""


import re


# Regular Expression to get the atoms in a reaction. Compile it once for all


def camel2underscores(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def varnames2ids(tmodel, variables):
    return [tmodel._var_dict[x].id for x in variables]